import pygame
import os


class Window:

    def __init__(self, size, title="Heat2D Game Project", icon=None, clear_color=(0, 0, 0)):
        self._width = size[0]
        self._height = size[1]
        self._size = size
        self._title = title
        self._icon = icon
        self.clear_color = clear_color

        if icon: self.icon = icon
        else: self.icon = os.path.join(__file__[:len(__file__) - len("window.py")], "favicon.png")
        self.size = size
        self.title = self._title

        self.center_x = self.width / 2
        self.center_y = self.height / 2
        self.center = (self.center_x, self.center_y)

        self.clock = pygame.time.Clock()
        self.max_fps = 60
        self.fps = self.max_fps

    def __repr__(self):
        return f"<heat2d.Window({self.size}, {self.title})>"

    @property
    def width(self): return self._width

    @width.setter
    def width(self, val):
        self._width = val
        self.surface = pygame.display.set_mode((self._width, self._height))

    @property
    def height(self): return self._height

    @height.setter
    def height(self, val):
        self._height = val
        self.surface = pygame.display.set_mode((self._width, self._height))

    @property
    def size(self): return self._size

    @size.setter
    def size(self, val):
        self._size = val
        self.surface = pygame.display.set_mode(self._size)

    @property
    def title(self): return self._title

    @title.setter
    def title(self, val):
        self._title = val
        pygame.display.set_caption(self._title)

    @property
    def icon(self): return self._icon

    @icon.setter
    def icon(self, val):
        self._icon = val
        pygame.display.set_icon(pygame.image.load(self._icon))



import pygame
from PIL import Image, ImageFilter

#Thanks to St. Paul for the optimization method for blurring

class GaussianBlur:

    def __init__(self, radius=3, resolution=40):
        self.radius = radius
        self.resolution = resolution

    def __repr__(self):
        return f"<heat2d.postprocess.GaussianBlur(radius={self.radius}, resolution={self.resolution})>"

    def process(self, display, size, position):
        surface = pygame.Surface(size)
        surface.blit(display, (0, 0), (position[0], position[1], size[0], size[1]))
        surface = pygame.transform.rotozoom(surface, 0, (self.resolution / 100.0) * 1)
        _size = surface.get_size()
        bytesurf = pygame.image.tostring(surface, "RGBA", False)
        bytesurf = Image.frombytes("RGBA", _size, bytesurf)
        bytesurf = bytesurf.filter(ImageFilter.GaussianBlur(radius=int(self.radius)))
        bytesurf = pygame.image.frombuffer(bytesurf.tobytes(), bytesurf.size, bytesurf.mode).convert()
        bytesurf = pygame.transform.rotozoom(bytesurf, 0, (100.0 / self.resolution) * 1)
        display.blit(bytesurf, position)


class BoxBlur:

    def __init__(self, radius=3, resolution=40):
        self.radius = radius
        self.resolution = resolution

    def __repr__(self):
        return f"<heat2d.postprocess.BoxBlur(radius={self.radius}, resolution={self.resolution})>"

    def process(self, display, size, position):
        surface = pygame.Surface(size)
        surface.blit(display, (0, 0), (position[0], position[1], size[0], size[1]))
        surface = pygame.transform.rotozoom(surface, 0, (self.resolution / 100.0) * 1)
        _size = surface.get_size()
        bytesurf = pygame.image.tostring(surface, "RGBA", False)
        bytesurf = Image.frombytes("RGBA", _size, bytesurf)
        bytesurf = bytesurf.filter(ImageFilter.BoxBlur(radius=int(self.radius)))
        bytesurf = pygame.image.frombuffer(bytesurf.tobytes(), bytesurf.size, bytesurf.mode).convert()
        bytesurf = pygame.transform.rotozoom(bytesurf, 0, (100.0 / self.resolution) * 1)
        display.blit(bytesurf, position)


class MedianFilter:

    def __init__(self, size=3, resolution=40):
        self.size = size
        self.resolution = resolution

    def __repr__(self):
        return f"<heat2d.postprocess.GaussianBlur(size={self.size}, resolution={self.resolution})>"

    def process(self, display, size, position):
        surface = pygame.Surface(size)
        surface.blit(display, (0, 0), (position[0], position[1], size[0], size[1]))
        surface = pygame.transform.rotozoom(surface, 0, (self.resolution / 100.0) * 1)
        _size = surface.get_size()
        bytesurf = pygame.image.tostring(surface, "RGBA", False)
        bytesurf = Image.frombytes("RGBA", _size, bytesurf)
        bytesurf = bytesurf.filter(ImageFilter.MedianFilter(size=int(self.size)))
        bytesurf = pygame.image.frombuffer(bytesurf.tobytes(), bytesurf.size, bytesurf.mode).convert()
        bytesurf = pygame.transform.rotozoom(bytesurf, 0, (100.0 / self.resolution) * 1)
        display.blit(bytesurf, position)


class MinFilter:

    def __init__(self, size=3, resolution=40):
        self.size = size
        self.resolution = resolution

    def __repr__(self):
        return f"<heat2d.postprocess.GaussianBlur(size={self.size}, resolution={self.resolution})>"

    def process(self, display, size, position):
        surface = pygame.Surface(size)
        surface.blit(display, (0, 0), (position[0], position[1], size[0], size[1]))
        surface = pygame.transform.rotozoom(surface, 0, (self.resolution / 100.0) * 1)
        _size = surface.get_size()
        bytesurf = pygame.image.tostring(surface, "RGBA", False)
        bytesurf = Image.frombytes("RGBA", _size, bytesurf)
        bytesurf = bytesurf.filter(ImageFilter.MinFilter(size=int(self.size)))
        bytesurf = pygame.image.frombuffer(bytesurf.tobytes(), bytesurf.size, bytesurf.mode).convert()
        bytesurf = pygame.transform.rotozoom(bytesurf, 0, (100.0 / self.resolution) * 1)
        display.blit(bytesurf, position)


class MaxFilter:

    def __init__(self, size=3, resolution=40):
        self.size = size
        self.resolution = resolution

    def __repr__(self):
        return f"<heat2d.postprocess.GaussianBlur(size={self.size}, resolution={self.resolution})>"

    def process(self, display, size, position):
        surface = pygame.Surface(size)
        surface.blit(display, (0, 0), (position[0], position[1], size[0], size[1]))
        surface = pygame.transform.rotozoom(surface, 0, (self.resolution / 100.0) * 1)
        _size = surface.get_size()
        bytesurf = pygame.image.tostring(surface, "RGBA", False)
        bytesurf = Image.frombytes("RGBA", _size, bytesurf)
        bytesurf = bytesurf.filter(ImageFilter.MaxFilter(size=int(self.size)))
        bytesurf = pygame.image.frombuffer(bytesurf.tobytes(), bytesurf.size, bytesurf.mode).convert()
        bytesurf = pygame.transform.rotozoom(bytesurf, 0, (100.0 / self.resolution) * 1)
        display.blit(bytesurf, position)


class ModeFilter:

    def __init__(self, size=3, resolution=40):
        self.size = size
        self.resolution = resolution

    def __repr__(self):
        return f"<heat2d.postprocess.GaussianBlur(size={self.size}, resolution={self.resolution})>"

    def process(self, display, size, position):
        surface = pygame.Surface(size)
        surface.blit(display, (0, 0), (position[0], position[1], size[0], size[1]))
        surface = pygame.transform.rotozoom(surface, 0, (self.resolution / 100.0) * 1)
        _size = surface.get_size()
        bytesurf = pygame.image.tostring(surface, "RGBA", False)
        bytesurf = Image.frombytes("RGBA", _size, bytesurf)
        bytesurf = bytesurf.filter(ImageFilter.ModeFilter(size=int(self.size)))
        bytesurf = pygame.image.frombuffer(bytesurf.tobytes(), bytesurf.size, bytesurf.mode).convert()
        bytesurf = pygame.transform.rotozoom(bytesurf, 0, (100.0 / self.resolution) * 1)
        display.blit(bytesurf, position)



import time
from heat2d import DISPATCHER


class Timer:

    def __init__(self, interval):
        self.init_time = time.time()
        self.interval = interval
        self.time = 0
        self.func = None

        DISPATCHER["engine"].timers.append(self)

    def __repr__(self):
        return f"heat2d.timer.Timer({self.time} ms)"

    def do(self, func):
        self.func = func

    def update(self):
        self.time = (time.time() - self.init_time) * 1000
        if self.time >= self.interval:
            self.time = 0
            if self.func: self.func()
            self.init_time = time.time()


class TickTimer:

    def __init__(self, interval):
        self.interval = interval
        self.tick = 0
        self.func = None

        DISPATCHER["engine"].timers.append(self)

    def __repr__(self):
        return f"heat2d.timer.TickTimer({self.tick} ticks)"

    def do(self, func):
        self.func = func

    def update(self):
        self.tick += 1
        if self.tick >= self.interval:
            self.tick = 0
            if self.func: self.func()



import pygame


class Context:

    def __init__(self, position, size):
        self.x, self.y = position
        self.width, self.height = size
        self.back_surface = pygame.Surface(size, pygame.SRCALPHA).convert_alpha()
        self.front_surface = pygame.Surface(size, pygame.SRCALPHA).convert_alpha()
        self.effect = None

    def __repr__(self):
        return f"<heat2d.ui.Context(({self.x},{self.y}), {self.width}x{self.height})>"

    def set_effect(self, effect):
        self.effect = effect



import pygame
pygame.font.init()
from heat2d import DISPATCHER
from heat2d.ui.context import Context
from heat2d.timer import Timer

def init():
    engine = DISPATCHER["engine"]

    fps_display = Context((0, 0), (40, 18))
    fps_display.back_surface.fill((0, 0, 0))
    fps_display.back_surface.set_alpha(128)

    engine.add(fps_display)

    font_calibri = pygame.font.SysFont("Calibri", 14)

    timer = Timer(300)

    @timer.do
    def function():
        fps_display.front_surface.fill((255, 255, 255, 0))
        fps_display.front_surface.blit(font_calibri.render(str(int(engine.window.fps)), True, (0, 255, 0)), (2, 2))



import pygame
from heat2d import DISPATCHER
from heat2d.window import Window
from heat2d.renderer import Renderer
from heat2d.stage import Stage
from heat2d.gameobject import GameObject
from heat2d.libs.keys import key_dictionary, button_dictionary, inv_key_dictionary, inv_button_dictionary
from heat2d.errors import NoStageDeclared
from heat2d.ui.context import Context
from heat2d import ui
from heat2d.timer import Timer, TickTimer


class Engine:

    def __init__(self):
        DISPATCHER["engine"] = self
        pygame.init()

        self.window = Window((800, 600))
        self.renderer = Renderer()

        self.timers = list()

        #working with a dictionary might be better than 3 lists
        self.key_states = {k: [0, 0, 0] for k in key_dictionary}
        self.mouse_states = {b: [0, 0, 0] for b in button_dictionary}
        self.mouse_x, self.mouse_y = 0, 0

        self.current_stage = None
        self.stages = dict()
        self.gameobjects = list()

        self.__is_running = False

        #Initialize modules
        ui.init()

    def __repr__(self):
        return f"<heat2d.Engine({self.window})>"

    #   @engine.input decorator function to store input calls
    def input(self, func):
        self.inp_funcs[func.__name__].append(func)

    #   @engine.event decorator function to store input calls
    def event(self, *args):
        def wrapper(func):
            self.event_funcs[func.__name__].append(func)
            if func.__name__ == "every_tick": self.timers.append(TickTimer(func, args[0]))
            elif func.__name__ == "every_time": self.timers.append(Timer(func, args[0]))

        return wrapper

    def remove_by_id(self, inpfunc, id):
        for func in self.inp_funcs[inp_func]:
            if id(func) == id:
                self.inp_funcs[inp_func].remove(func)
                return

    #   @engine.add decorator to add components into engine
    def add(self, comp):
        self._add(comp)
        #if any(tuple(parent == Stage for parent in cls.__bases__)):
        #    self.add(cls)
        #else:
        #    raise ValueError(f"{cls} is not a Stage")

    def _add(self, *args):
        for arg in args:
            if isinstance(arg, Context):
                self.renderer.ui_layers.append(arg)

            else:
                #Initialize objects
                inst = arg()

                if isinstance(inst, Stage):
                    self.stages[inst.__class__.__name__] = inst
                    if len(self.stages) == 1: self.current_stage = inst.__class__.__name__

                elif isinstance(inst, GameObject):
                    self.gameobjects.append(inst)

    def change_stage(self, stage):
        if stage in self.stages: self.current_stage = stage
        else: raise NameError(f"{stage} is not found on stages.")

    def key_pressed(self, key):
        if self.key_states[key][1]: return True
        return False

    def key_released(self, key):
        if self.key_states[key][2]: return True
        return False

    def key_held(self, key):
        if self.key_states[key][0]: return True
        return False

    def mouse_pressed(self, button):
        if self.mouse_states[button][1]: return True
        return False

    def mouse_released(self, button):
        if self.mouse_states[button][2]: return True
        return False

    def mouse_held(self, button):
        if self.mouse_states[button][0]: return True
        return False

    def mouse_wheel_up(self):
        if self.mouse_states["wheelup"][1]: return True
        return False

    def mouse_wheel_down(self):
        if self.mouse_states["wheeldown"][1]: return True
        return False

    def handle_events(self):
        self.mouse_x, self.mouse_y = pygame.mouse.get_pos()
        for k in self.key_states:
            self.key_states[k][1] = 0
            self.key_states[k][2] = 0

        for b in self.mouse_states:
            self.mouse_states[b][1] = 0
            self.mouse_states[b][2] = 0

        self.events = pygame.event.get()
        for event in self.events:
            if event.type == pygame.QUIT: self.__is_running = False

            elif event.type == pygame.KEYDOWN:
                self.key_states[inv_key_dictionary[event.key]][0] = 1
                self.key_states[inv_key_dictionary[event.key]][1] = 1

            elif event.type == pygame.KEYUP:
                self.key_states[inv_key_dictionary[event.key]][0] = 0
                self.key_states[inv_key_dictionary[event.key]][1] = 0
                self.key_states[inv_key_dictionary[event.key]][2] = 1

            elif event.type == pygame.MOUSEBUTTONDOWN:
                self.mouse_states[inv_button_dictionary[event.button]][0] = 1
                self.mouse_states[inv_button_dictionary[event.button]][1] = 1

            elif event.type == pygame.MOUSEBUTTONUP:
                self.mouse_states[inv_button_dictionary[event.button]][0] = 0
                self.mouse_states[inv_button_dictionary[event.button]][1] = 0
                self.mouse_states[inv_button_dictionary[event.button]][2] = 1

    def run(self):
        self.__is_running = True

        if not self.current_stage: raise NoStageDeclared("You have to declare at least one stage to run the engine.")

        while self.__is_running:
            self.window.clock.tick(self.window.max_fps)
            self.window.fps = self.window.clock.get_fps()

            self.handle_events()
            self.stages[self.current_stage].update()

            for timer in self.timers: timer.update()
            self.renderer.draw()

        pygame.quit()



"""
      HEAT2D Game Engine
     GPL 3.0 © Kadir Aksoy
https://github.com/kadir014/heat2d

"""

VERSION =       "0.0.2a1"
VERSION_TUPLE = (0, 0, 2)
VERSION_STATE = "alpha"
LICENSE =       "GNU General Public License v3.0"

import sys
PLATFORM = None
if   sys.platform == "win32": PLATFORM = "Windows"
elif sys.platform == "darwin":  PLATFORM = "MacOS"
elif sys.platform.startswith("linux"): PLATFORM = "Linux"
elif sys.platform.startswith("freebsd"): PLATFORM = "FreeBSD"

import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "1"
import pygame

DISPATCHER = dict()
THREADS = list()
TOTAL_THREADS = 0

if "HEAT2D_DISABLE_ANSI_COLORS" in os.environ: __USE_ANSI_COLORS = False
else: __USE_ANSI_COLORS = True

from heat2d.libs.terminal import REVERSE, RESET, FG
if "HEAT2D_HIDE_WELCOME_MESSAGE" not in os.environ:
    if __USE_ANSI_COLORS:
        print(f"\u2588{REVERSE}" + "\u2584"*20 + f"{RESET}\u2588")
        print(f"\u2588{FG.lightred}       HEAT2D       {RESET}\u2588 Heat2D Version: 0.0.0")
        print(f"\u2588{FG.orange}    GAME ENGINE     {RESET}\u2588 Pygame Version: {pygame.version.ver}")
        print("\u2588" + "\u2584"*20 + "\u2588\n")
    else:
        print(f"\u2588{REVERSE}" + "\u2584"*20 + f"{RESET}\u2588")
        print(f"\u2588       HEAT2D       \u2588 Heat2D Version: 0.0.0")
        print(f"\u2588    GAME ENGINE     \u2588 Pygame Version: {pygame.version.ver}")
        print("\u2588" + "\u2584"*20 + "\u2588\n")

from heat2d.engine import Engine
from heat2d.stage import Stage
from heat2d.gameobject import GameObject
from heat2d.sprite import Sprite
from heat2d import visuals
from heat2d.timer import Timer, TickTimer
from heat2d import ui
from heat2d import postprocess
from heat2d import networking

del sys, os, pygame, libs, engine, errors, window, gameobject, renderer, sprite, stage, timer, REVERSE, RESET, FG



import pygame
from heat2d import DISPATCHER


class Renderer:

    def __init__(self):
        self.window = DISPATCHER["engine"].window
        self.event_funcs = {"draw_last" : list(), "draw_first" : list()}

        self.ui_layers = list()

    def __repr__(self):
        return "<heat2d.Renderer>"

    def event(self, func):
        self.event_funcs[func.__name__].append(func)

    def draw(self):
        self.window.surface.fill(self.window.clear_color)

        gameobjects = DISPATCHER["engine"].stages[DISPATCHER["engine"].current_stage].gameobjects
        for gameobject in gameobjects:
            gameobject.update()
            self.window.surface.blit(gameobject.sprite.surface, (gameobject.x - gameobject.sprite.surface_width/2, gameobject.y - gameobject.sprite.surface_height/2))

        for layer in self.ui_layers:
            if layer.effect:
                layer.effect.process(self.window.surface, layer.back_surface.get_size(), (layer.x, layer.y))
            self.window.surface.blit(layer.back_surface, (layer.x, layer.y))
            self.window.surface.blit(layer.front_surface, (layer.x, layer.y))

        pygame.display.flip()



import pygame
from math import cos, sin, pi, atan2, degrees
from heat2d.visuals import Rectangle


class Sprite:
    def __init__(self, filename):
        if isinstance(filename, Rectangle):
            self.__source_image = pygame.Surface((filename.width, filename.height)).convert()
            self.__source_image.fill(filename.color)
        else:
            self.__source_image = pygame.image.load(filename)

        self.__calc(self.__source_image)

        self.angle = 0

        self.__blit(False)

    def __calc(self, source_surface):
        self.source_image = source_surface.convert_alpha()
        self.source_size = source_surface.get_size()
        self.width, self.height = self.source_size

        if self.width != self.height:
            max_size = max(self.source_size)

            temp_surf = pygame.Surface((max_size, max_size), pygame.SRCALPHA)
            temp_surf.blit(self.source_image, (max_size / 2 - self.width / 2,
                                               max_size / 2 - self.height / 2))

            self.__calc(temp_surf)
            return

        self.width_constant = abs(self.width - (cos(pi / 4) * self.width) * 2)
        self.height_constant = abs(self.height - (sin(pi / 4) * self.height) * 2)

        self.surface = pygame.Surface((self.width + self.width_constant,
                                       self.height + self.height_constant),
                                       pygame.SRCALPHA).convert_alpha()

        self.size = self.surface.get_size()
        self.surface_width, self.surface_height = self.size

        self.flip_horizontal = False
        self.flip_vertical = False

    def __blit(self, rotated=True):
        self.surface.fill((255, 255, 255, 0))

        if rotated:
            self.rotated_surface = pygame.transform.rotate(self.source_image, self.angle)

            self.surface.blit(self.rotated_surface, (-self.rotated_surface.get_width() / 2 + (self.width + self.width_constant) / 2,
                                                     -self.rotated_surface.get_height() / 2 + (self.height + self.height_constant) / 2))

        else:
            self.surface.blit(self.source_image, (-self.width / 2 + (self.width + self.width_constant) / 2,
                                                  -self.height / 2 + (self.height + self.height_constant) / 2))

    def get_hitbox(self):
        return (-self.rotated_surface.get_width() / 2 + (self.width + self.width_constant) / 2,
                -self.rotated_surface.get_height() / 2 + (self.height + self.height_constant) / 2,
                self.rotated_surface.get_width(), self.rotated_surface.get_height())

    def scale(self, scale):
        self.__calc(pygame.transform.scale(self.__source_image, (self.width * scale, self.height * scale)))
        self.__blit(False)

    def set_size(self, size):
        self.__calc(pygame.transform.scale(self.__source_image, size))
        self.__blit(False)

    def rotate(self, angle):
        self.angle = (self.angle + angle) % 360
        self.__blit()

    def rotate_to(self, x, y, tx, ty):
        self.angle = (-degrees(atan2(y - ty,  x - tx)) - 180) % 360
        self.__blit()

    def flip(self, horizontal=False, vertical=False):
        if horizontal:
            if self.flip_horizontal: self.flip_horizontal = False
            else: self.flip_horizontal = True
            self.source_image = pygame.transform.flip(self.source_image, True, False)
            self.__blit(False)

        elif vertical:
            if self.flip_vertical: self.flip_vertical = False
            else: self.flip_vertical = True
            self.source_image = pygame.transform.flip(self.source_image, False, True)
            self.__blit(False)

    def orient(self, orientation):
        if orientation.lower() == "east" and self.flip_horizontal:
            self.flip(horizontal=True)

        elif orientation.lower() == "west" and not self.flip_horizontal:
            self.flip(horizontal=True)

        elif orientation.lower() == "north" and self.flip_vertical:
            self.flip(vertical=True)

        elif orientation.lower() == "south" and not self.flip_vertical:
            self.flip(vertical=True)



class Stage:

    def __init__(self):
        self.gameobjects = list()

    def __repr__(self):
        return f"<heat2d.Stage({self.__class__.__name__})>"

    def add(self, gameobject):
        self.gameobjects.append(gameobject())

    def update(self):
        pass



#from distutils.core import setup
from setuptools import setup, find_packages

setup(
    name="heat2d",
    packages = find_packages(),
    version="0.0.2a1",
    author="Kadir Aksoy",
    author_email="kursatkadir014@gmail.com",
    description="A simple 2D game engine",
    url="https://github.com/kadir014/heat2d",
    project_urls={
    'Documentation': 'https://kadir014.github.io/projects/heat2d/index.html',
    },
    keywords='game engine development 2d',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=["pygame>=1.9.6"]
)



from heat2d import __USE_ANSI_COLORS
from heat2d.libs.terminal import RESET, FG


def warn(message):
    if __USE_ANSI_COLORS: print(f"{FG.orange}WARNING:{RESET} {message}")
    else: print(f"WARNING: {message}")

class NoStageDeclared(Exception): pass
class NetworkingError(Exception): pass



import socket
import threading
import time
from heat2d.errors import NetworkingError

HOSTNAME  = socket.gethostname()
LOCALHOST = socket.gethostbyname(HOSTNAME)


class Header:

    def __init__(self, string):
        i = 1
        s = ""
        self.fields = list()

        while i < len(string)+1:
            if string[i] == "}": break
            if string[i] == "/":
                self.fields.append(s)
                continue
            s += string[i]

    def __repr__(self):
        return f"<heat2d.networking.Header({self.fields})>"


class ConnectedClient(threading.Thread):

    def __init__(self, s, address, id, server):
        super().__init__()
        self.socket = s
        self.address = address[0]
        self.port = address[1]
        self.id = id
        self.server = server
        self.__is_alive  = True

        self.package = ""
        self.header = None

    def __repr__(self):
        return f"<heat2d.networking.ConnectedClient(id={self.id}, {self.address}:{self.port})>"

    def run(self):
        print(f"New client has connected {self}")

        while self.__is_alive :
            s = self.socket.recv(1024).decode("utf-8")

            if not self.header:
                self.header = Header(s[:s.find("}")])
                self.package += s[9:]

            else:
                if len(self.package) > int(self.header.fields[0]):
                    self.package = ""
                    self.header = None
                    if self.header.fields[1] == "str":
                        self.server.callback("string_received", str, self)
                        self.socket.send("{0/ping}".encode("utf-8"))
                    continue
                self.package += s

            #if len(str) > 0:
            #    self.server.callback("message_received", str, self)
            #    self.socket.send(f"Message '{str}' received succesfully.".encode("utf-8"))


class Server(threading.Thread):

    def __init__(self, address, port):
        super().__init__()
        self.address, self.port = address, port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.clients = list()
        self.event_funcs = dict()
        self.__is_alive = True

    def __repr__(self):
        return f"<heat2d.networking.Server({self.address}:{self.port})>"

    def event(self, func):
        self.event_funcs[func.__name__] = func

    def callback(self, eventname, *args, **kwargs):
        if eventname in self.event_funcs: self.event_funcs[eventname](*args, **kwargs)

    def run(self):
        self.socket.bind((self.address, self.port))
        self.socket.listen(5)

        print(f"Server started listening on {self.address}:{self.port}\n")

        while self.__is_alive:
            connection = self.socket.accept()
            connectedclient = ConnectedClient(connection[0], connection[1], len(self.clients), self)
            connectedclient.start()
            self.clients.append(connectedclient)


class Connection(threading.Thread):

    def __init__(self, address, port):
        super().__init__()
        self.address, self.port = address, port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.event_funcs = dict()
        self.latency = 0
        self.l_start = 0
        self.__is_alive  = True

    def __repr__(self):
        return f"<heat2d.networking.Connection({self.address}:{self.port})>"

    def event(self, func):
        self.event_funcs[func.__name__] = func

    def callback(self, eventname, *args, **kwargs):
        if eventname in self.event_funcs: self.event_funcs[eventname](*args, **kwargs)

    def send_str(self, s):
        self.l_start = time.time()
        self.socket.send(f"{{{len(s)}/str}}{s}".encode("utf-8"))

    def run(self):
        self.socket.connect((self.address, self.port))
        print(f"Connection established with {self.address}:{self.port}")

        while self.__is_alive :
            s = self.socket.recv(1024).decode("utf-8")

            if not self.header:
                self.header = Header(s[:s.find("}")])
                self.package += s[9:]

            else:
                if len(self.package) > int(self.header.fields[0]):
                    self.package = ""
                    self.header = None
                    if self.header.fields[1] == "ping": self.latency = time.time() - self.start_l
                    continue
                self.package += s

        # send server : client disconnected
        self.socket.close()


def initialize_server(address, port=1234):
    if not address: raise NetworkingError("No address has been passed.")
    return Server(address, port)

def create_connection(address, port=1234):
    if not address: raise NetworkingError("No address has been passed.")
    conn = Connection(address, port)
    conn.start()
    return conn
